/******************************************************************** 
 * Quantum Hierarchical State Machine declarations (C version)
 * Copyright 2002-2003 by Miro Samek, http://www.quantum-leaps.com 
 * All Rights Reserved.
 *******************************************************************/
#ifndef qhsm_h
#define qhsm_h

#ifndef qevent_h
   #include "qevent.h"
#endif

typedef struct QHsm QHsm;
typedef void (*QPseudoState)(QHsm *, QEvent const *);
typedef QPseudoState (*QState)(QHsm *, QEvent const *);
typedef QPseudoState QSTATE;  /* return value from a state-handler */

struct QHsm {             /* Hierarchical State Machine base class */
   QState state__;                             /* the active state */
   QState source__;             /* source state during a transiton */
};

/* public members */
void QHsmInit(QHsm *me, QEvent const *e);    /* initial transition */
void QHsmDispatch(QHsm *me, QEvent const *e);     /* take RTC step */
                                /* static method (no "me" pointer) */
char const *QHsmGetVersion(void);

/* protected members */
QHsm *QHsmCtor_(QHsm *me, QPseudoState initial);           /* Ctor */
void QHsmXtor_(QHsm *me);                                  /* Xtor */

QSTATE QHsm_top(QHsm *me, QEvent const *);          /* "top" state */
#define QHsmGetState_(me_) ((me_)->state__)
void QHsmTran_(QHsm *me, QState target);     /* dynamic transition */
#define Q_INIT(target_) (((QHsm*)me)->state__ = (QState)(target_))
#define Q_TRAN(target_) \
   QHsmTran_((QHsm *)me, (QState)(target_))

    /* helper macro to calculate static dimension of a 1-dim array */
#define DIM(array_) (sizeof(array_)/sizeof(array_[0]))

#endif                                                   /* qhsm_h */
