/////////////////////////////////////////////////////////////////////
// Quantum Hierarchical State Machine declaration (C++ version)
// Copyright 2002-2003 by Miro Samek, http://www.quantum-leaps.com 
// All Rights Reserved.
/////////////////////////////////////////////////////////////////////
#ifndef qhsm_h
#define qhsm_h

#ifndef qevent_h
   #include "qevent.h"
#endif

class QHsm {                    // Quantum Hierarchical State Machine
public:
   typedef void (QHsm::*QPseudoState)(QEvent const*);  // pseudostate
   typedef QPseudoState (QHsm::*QState)(QEvent const*);      // state

   void init(QEvent const *e = 0);      // execute initial transition
   void dispatch(QEvent const *e);                  // dispatch event
   static char const *getVersion();

protected:
   QHsm(QPseudoState initial);                                // ctor
   virtual ~QHsm();                                   // virtual xtor
   QPseudoState top(QEvent const*) { return 0; }   // the "top" state
   QState getState() const { return myState; }
   void tran(QState target);              // dynamic state transition
   void init_(QState target) { myState = target; }
   #define Q_INIT(target_) init_(Q_STATIC_CAST(QState, target_))
   #define Q_TRAN(target_) tran(Q_STATIC_CAST(QState, target_))

private:
   QState myState;                                // the active state
   QState mySource;               // source state during a transition
};
typedef QHsm::QPseudoState QSTATE;       // state-handler return type

                                         // compiler-specific casting
#define Q_STATIC_CAST(type_, expr_) static_cast<type_>(expr_)
#define Q_STATE_CAST(x_)  reinterpret_cast<QState>(x_)
       // helper macro to calculate static dimension of a 1-dim array
#define DIM(array_) (sizeof(array_)/sizeof(array_[0]))

#endif                                                      // qhsm_h
