#ifndef SEGSET_H
#define SEGSET_H

#include <set>
#include "segment.h"

class MutableSegmentSet;

class fCompare
{
    public:
        fCompare(const MutableSegmentSet& set) : m_set(set) {}
        
        bool operator() (const Segment& lhs, const Segment& rhs) const;
        
    private:
        const MutableSegmentSet& m_set;
};

class MutableSegmentSet
{
    public:
        typedef std::set<Segment, fCompare> SegSet;
        typedef SegSet::iterator iterator;
        typedef SegSet::const_iterator const_iterator;

        MutableSegmentSet() : m_set(fCompare(*this)), m_scan_pos(0) {}
        
        void set_scan_pos(int x)
        {
            m_scan_pos = x;
            reorder();
        }
        int get_scan_pos() const
        {
            return m_scan_pos;
        }
        
        iterator       begin()       { return m_set.begin(); }
        iterator       end()         { return m_set.end();   }
        const_iterator begin() const { return m_set.begin(); }
        const_iterator end()   const { return m_set.end();   }

        std::pair<iterator, bool> insert(const Segment& s)
        { 
            return m_set.insert(s);
        }
        void erase(iterator i)
        { 
            m_set.erase(i);
        }
        iterator find(const Segment& s) const
        { 
            return m_set.find(s);
        }
    private:
        std::set<Segment, fCompare> m_set;
        int m_scan_pos;
        
        class MutableIter : public SegSet::iterator
        {
            public:
                typedef SegSet::iterator BaseIter;

                MutableIter() {}
                MutableIter(BaseIter bi) : BaseIter(bi) {}
                
                Segment& operator*()
                {
                    return const_cast<Segment&>(this->BaseIter::operator*());
                }
                const Segment& operator*() const
                {
                    return this->BaseIter::operator*();
                }
                MutableIter& operator++()
                {
                    this->BaseIter::operator++();
                    return *this;
                }
                const MutableIter operator++(int)
                {
                    MutableIter temp = *this;
                    ++(*this);
                    return temp;
                }
                MutableIter& operator--()
                {
                    this->BaseIter::operator--();
                    return *this;
                }
                const MutableIter operator--(int)
                {
                    MutableIter temp = *this;
                    --(*this);
                    return temp;
                }
        };

        void reorder();
        void reverse(iterator first, iterator last);
        iterator get_intersecting_segments(iterator first, iterator last);
};

inline bool fCompare::operator() (const Segment& lhs, const Segment& rhs) const
{
    int x_coord = m_set.get_scan_pos();

    if (lhs.evaluate(x_coord) < rhs.evaluate(x_coord))
    {
        return true;
    }
    else if (lhs.evaluate(x_coord) == rhs.evaluate(x_coord))
    {
        if (lhs.slope() < rhs.slope())
        {
            return true;
        }
        else if (lhs.slope() == rhs.slope())
        {
            if (lhs.LowerX() < rhs.LowerX())
            {
                return true;
            }
            else if (lhs.LowerX() == rhs.LowerX())
            {
                if (lhs.UpperX() < rhs.UpperX())
                {
                    return true;
                }
            }
        }
    }
    return false;
}

#endif
